#!/bin/bash
# Script to capture set of debug information
# This script will move to /tmp, create a temporary working directory and capture information about the Graphcore tools and system environment
# It requires the Graphcore tools to be in your path, and will query for sudo priviliges if you aren't root
# The resulting compressed log file is typically <1MB
####################################################
############    INFORMATION CAPTURED ###############
# gc-info -l
# gc-inventory
# gc-links -j -v
# dkms ipu_driver
# modinfo ipu_driver
# Where available /etc/os-release, /etc/redhat-release, /etc/debian_version
# dmidecode
# lspci -vvv
# Where available /var/log/messages, /var/log/syslog, /var/log/kern.log
# dmesg
####################################################


# trap interrupts to allow us to cleanup
trap cleanup INT

function cleanup {
# delete my temporary directory if I get interrupted
if [ -d $targetdir ]; then
    echo " "
    echo "Interrupt - cleaning up temporary directory $PWD/$targetdir"
    rm -r $targetdir
fi

exit 1

}


# helper functions

function grab_file_su {

file=$1
target=$2

if [ -f "$1" ]; then
    sudo cat $1 > $targetdir/$target
else
    echo "$1 not found" > $targetdir/gc_capture_script.log
fi

}

function grab_file {

file=$1
target=$2

if [ -f "$1" ]; then
    cat $1 > $targetdir/${target}.out
else
    echo "$1 not found" > $targetdir/gc_capture_script.log
fi

}


# check whether we have the drivers loaded
if ! [ -x "$(command -v gc-info)" ]; then
    echo 'Error: gc-tools (e.g. gc-info etc.) are not in path - please source the gc_drivers enable.sh script.' >&2
    exit 1
fi

# change our working directory to /tmp/
cd /tmp

# set up a consistent directory for this run of the debug
timestamp=$(date +"%Y-%m-%d_%H-%M-%S")
targetdir=gc_debug_${timestamp}
mkdir -p $targetdir
echo "Created working directory $PWD/$targetdir"

############### Capture output of Graphcore tools
echo -n "Capturing Graphcore tool output . . . gc-info"
# capture gc-info
gc-info -l > $targetdir/gc-info.out

echo -n " . . . gc-inventory"
# grab gc-inventory
gc-inventory > $targetdir/gc-inventory.out

echo -n " . . . gc-links (~20 secs)"
# grab gc-links
gc-links -j -v > $targetdir/gc-links.out
echo " . . . done"

############# Capture Graphcore driver info
echo -n "Capturing Graphcore driver info"
dkms ipu_driver status > $targetdir/dkms_ipudriver.out

modinfo ipu_driver > $targetdir/modinfo_ipudriver.out
echo " . . . done"


############# Capture Generic Linux information
echo -n "Capturing Linux system information . . . requires SUDO . . . "

uname -a > $targetdir/uname.out

grab_file /etc/os-release os-release.out
grab_file /etc/redhat-release redhat-release.out
grab_file /etc/debian_version debian_version.out

sudo dmidecode > $targetdir/dmidecode.out

sudo lspci -vvv > $targetdir/lspci-vvv.out

grab_file_su /var/log/messages var-log-messages.out
grab_file_su /var/log/syslog var-log-syslog.out
grab_file_su /var/log/kern.log var-log-kernlog.out

dmesg > $targetdir/dmesg.out
echo "done"

# create our output tgz
echo -n "Creating output tgz $PWD/${targetdir}.tgz"
tar -czf ${targetdir}.tgz "$targetdir"
echo " . . . done"

# cleaning up my temporary directory
echo -n "Cleaning up temporary directory $PWD/${targetdir}"
rm -r "$targetdir"
echo " . . . done"

echo "Please return $PWD/${targetdir}.tgz to your Graphcore support representative"
