#ifndef Compression_hpp
#define Compression_hpp

#include <cstdint>
#include <lz4frame.h>
#include <memory>
#include <stdexcept>
#include <string>

// Compress data using LZ4 default settings. Throws on error (which should
// basically never happen).
template <typename T>
T compressLZ4(const char *data, std::size_t size) {
  using namespace std::string_literals;

  static_assert(sizeof(*std::declval<T>().data()) == 1,
                "The elements of T must be 1 byte");

  LZ4F_preferences_t prefs{
    {
      // Block size.
      LZ4F_max1MB,
      // Whether blocks depend on previous ones.
      LZ4F_blockIndependent,
      // Checksum at the end of the frame.
      LZ4F_contentChecksumEnabled,
      // Frame type.
      LZ4F_frame,
      // Content size.
      size,
      // No dictionary.
      0,
      // No block checksum.
      LZ4F_noBlockChecksum
    },
    // Fastest compression.
    0,
    // No autoflush.
    0,
    // Favour decompression speed (only relevant for high compression mode).
    0,
    // Reserved.
    { 0, 0, 0 },
  };

  std::size_t dstCapacity = LZ4F_compressFrameBound(size, &prefs);

  T dst;
  dst.resize(dstCapacity);
  // TODO: The const cast can be removed in C++17.
  std::size_t ret = LZ4F_compressFrame(
    const_cast<void*>(static_cast<const void*>(dst.data())),
    dst.size(),
    data,
    size,
    &prefs
  );

  if (LZ4F_isError(ret)) {
    throw std::runtime_error("error compressing LZ4 frame: "s +
                             LZ4F_getErrorName(ret));
  }

  dst.resize(ret);
  return dst;
}

// Decompress LZ4 data to a string or vector<char>. Throws on error.
// Decompressing to a string is useful because elfio can read from an istream
// but not from an array, and making an istream from a vector is non-trivial
// (see VectorStreamBuf in Poplar). If the output size is bigger than
// maxOutSize, and maxOutSize is not 0, then an exception is thrown.
template <typename T>
T decompressLZ4(const char *data, std::size_t size, std::size_t maxOutSize) {
  using namespace std::string_literals;

  static_assert(sizeof(*std::declval<T>().data()) == 1,
                "The elements of T must be 1 byte");

  LZ4F_decompressionContext_t ctx = nullptr;
  std::size_t ret = LZ4F_createDecompressionContext(&ctx, LZ4F_VERSION);
  if (LZ4F_isError(ret)) {
    throw std::runtime_error("error creating LZ4 decompression context: "s +
                             LZ4F_getErrorName(ret));
  }

  // Automatically free the context when leaving this scope.
  struct FreeContext {
    LZ4F_decompressionContext_t ctx;
    ~FreeContext() {
      LZ4F_freeDecompressionContext(ctx);
    }
  } ctxFreer{ctx};

  // Get the decompressed size.
  LZ4F_frameInfo_t frameInfo{};

  std::size_t frameHeaderSize = size;
  ret = LZ4F_getFrameInfo(ctx, &frameInfo, data, &frameHeaderSize);
  if (LZ4F_isError(ret)) {
    throw std::runtime_error("error getting LZ4 frame info: "s +
                             LZ4F_getErrorName(ret));
  }

  // Check that the output size isn't bigger than the limit given. Note
  // that frameInfo.contentSize == 0 means "unknown" so we also have to check
  // in the loop below.
  if (maxOutSize != 0 && frameInfo.contentSize > maxOutSize) {
    throw std::runtime_error("decompressed size larger than limit: " +
                             std::to_string(frameInfo.contentSize) + " > " +
                             std::to_string(maxOutSize));
  }

  const std::size_t inputBlockSize = 64 * 1024;

  // Unfortunately we cannot rely on `frameInfo.contentSize`. It has a special
  // value of 0 which means "unknown", and although we try and set it to
  // a non-zero value in every case, we can't do that if the content size is
  // actually zero. So we use it as a hint to reserve space in dst, and
  // do the decompression in a loop. I used `resize()` rather than `reserve()`
  // to reduce the number of methods required on `T`.
  T dst;
  dst.resize(frameInfo.contentSize);

  std::size_t srcIdx = frameHeaderSize;
  std::size_t dstIdx = 0;

  while (srcIdx < size) {
    if (dstIdx + inputBlockSize > dst.size()) {
      dst.resize(dstIdx + inputBlockSize);
    }

    std::size_t dstSize = dst.size() - dstIdx;
    std::size_t srcSize = size - srcIdx;
    // TODO: The const cast can be removed in C++17.
    ret = LZ4F_decompress(
      ctx,
      const_cast<void*>(static_cast<const void*>(dst.data() + dstIdx)),
      &dstSize,
      data + srcIdx,
      &srcSize,
      nullptr
    );

    if (LZ4F_isError(ret)) {
      throw std::runtime_error("error decompressing LZ4 frame: "s +
                              LZ4F_getErrorName(ret));
    }

    srcIdx += srcSize;
    dstIdx += dstSize;

    if (maxOutSize != 0 && dstIdx > maxOutSize) {
      throw std::runtime_error("decompressed size larger than limit: " +
                               std::to_string(maxOutSize));
    }
  }

  dst.resize(dstIdx);

  // Check the frameInfo content size matches the actual size (unless it is
  // 0 which means "unknown").
  if (frameInfo.contentSize != 0 && dst.size() != frameInfo.contentSize) {
    throw std::runtime_error("error decompressing LZ4 frame: read " +
                             std::to_string(dst.size()) +
                             " bytes but expected " +
                             std::to_string(frameInfo.contentSize));
  }

  return dst;
}

#endif // Compression_hpp
