/*
 * Copyright (c) 2021 Graphcore Ltd, All rights reserved.
 */

#ifndef _GCIPUINFO_H
#define _GCIPUINFO_H

#include <map>
#include <stdexcept>
#include <string>
#include <vector>

enum DeviceDiscoveryMode {
  /**
   * Discover all devices in current partition.
   * This is the default mode
   */
  DiscoverActivePartitionIPUs,
  /**
   * Discover all devices across all partitions.
   */
  DiscoverAllPartitionIPUs,
};

class gcipuinfo {
private:
  std::vector<std::map<std::string, std::string>> data;
  bool autoUpdateMode = true;

public:
  gcipuinfo(DeviceDiscoveryMode = DiscoverActivePartitionIPUs);

  ///@{
  /**
   * @name Device attribute query methods.
   */
  /** Updates the attribute info.
   *
   *  Calling this function updates the data. Alternatively, you can call
   *  \ref setUpdateMode() to ensure the data is always updated.
   *
   *  \return `true` if the device attributes have been successfully updated,
   *  `false` if an error occurred.
   */
  bool updateData();

  /** Sets the attribute update mode.
   *
   *  By default, the attribute data is queried on
   *  construction of this class. This function selects between the default
   *  behaviour, and always updating on API call.
   *
   *  \param autoUpdate `true` to enable auto-update mode, `false` to use the
   *                    original data.
   */
  void setUpdateMode(bool autoUpdate);

  /** Get all attributes, for all devices.
   *
   *  \return A `std::vector` containing a `std::map` for each device.
   */
  std::vector<std::map<std::string, std::string>> getDevices();

  /** Return a JSON representation of all device attributes.
   *
   *  \return A JSON-formatted tree of devices
   *  and device attributes as a `std::string`
   */
  std::string getDevicesAsJSON();

  /** Get all attributes, for a specific devices.
   *
   *  \param deviceId Device ID to query.
   *  \return A `std::map` for each device.
   */
  std::map<std::string, std::string> getAttributesForDevice(unsigned deviceId);

  /** Get a specific attribute for a specific device.
   *
   *  \param deviceId Device ID to query.
   *  \param key Device attribute name.
   *
   *  \return The attribute.
   */
  std::string getNamedAttributeForDevice(unsigned deviceId,
                                         const std::string key);

  /** Get a specific attribute for all devices.
   *
   *  \param key Device attribute name.
   *  \return A `std::vector` containing the attribute for all devices.
   */
  std::vector<std::string> getNamedAttributeForAll(const std::string key);
  ///@}

  /*
   * Application Event Record
   *
   * Application events are recorded from within GCDA, Poplar or frameworks
   * to indicate some kind of issue with the current application.
   */

  /** Application Event Record severity level.
   *
   * The severity level of an application event indicates how serious it is
   * and potential ways of resolving the issue.
   */
  enum EventSeverity {
    EventSevNone = 0,

    /**
     * An event which may indicate a system problem.
     *
     * text: "warning"
     */
    EventSevWarning = 1,

    /**
     * An error likely in the application code or configuration.
     *
     * poplar::poplar_error, poplar::application_runtime_error
     *
     * text: "application"
     */
    EventSevApplicationError = 2,

    /**
     * It is not known if this is a system error, or an error
     * in the application code or configuration.
     *
     * poplar::unknown_runtime_error
     *
     * text: "undetermined"
     */
    EventSevUndeterminedError = 3,

    /**
     * The error may be resolvable by an IPU reset or a partition reset (for
     * POD systems) or a link reset (for non-Pod systems).
     *
     * poplar::recoverable_runtime_error + IPU_RESET or PARTITION_RESET or
     * LINK_RESET
     *
     * text: "requires_user_reset"
     */
    EventSevRequiresUserReset = 4,

    /**
     * The error may be resolvable by a full reboot of the IPU-M
     * system or Poplar server.
     *
     * poplar::recoverable_runtime_error + FULL_RESET
     *
     * text: "requires_system_reset"
     */
    EventSevRequiresSystemReset = 5,

    /**
     * The error may require admin-level system reconfiguration
     * or hardware replacement.
     *
     * poplar::unrecoverable_runtime_error
     *
     * text: "nonrecoverable"
     */
    EventSevNonRecoverable = 6
  };

  ///@{
  /**
   * @name Application Event Record key names
   */
  /// The path the record was stored at
  static constexpr const char *keyRecordPath = "event record path";
  /// When the event was recorded
  static constexpr const char *keyTimestamp = "timestamp";
  /// Severity level - \ref EventSeverity
  static constexpr const char *keySeverity = "severity";
  /// Command line of the process that recorded the event
  static constexpr const char *keyCommandLine = "command line";
  /// Process id of the process that recorded the event
  static constexpr const char *keyPid = "pid";
  /// List of any attached IPUs (device ids)
  static constexpr const char *keyAttachedIPUs = "attached ipus";
  /// List of any specific IPUs named in this event (device ids)
  static constexpr const char *keySpecificIPUs = "specific ipus";
  /// List of all currently attached IPU-Machine hostnames
  static constexpr const char *keyAttachedIPUHosts = "attached ipu hosts";
  /// List of any IPU-Machine hostnames associated with
  /// the devices named in "specific ipus"
  static constexpr const char *keySpecificIPUHosts = "specific ipu hosts";
  /// The partition in use by the application, if applicable
  static constexpr const char *keyPartition = "partition";
  /// A description of the event
  static constexpr const char *keyDescription = "description";
  ///@}

  ///@{
  /**
   * @name Application Event Record query methods.
   */
  /**
   * Return a JSON-formatted string describing the last recorded
   * application event.  If there is no recorded event, or the
   * event has a \ref EventSeverity below \ref minimumSeverity,
   * an empty JSON dictionary "{}" is returned.
   *
   * Looks in the path specified by \param eventRecordPath
   * for an event record.  If this is empty, falls back to
   * the path in `IPU_APP_EVENT_RECORD_PATH`.  If neither are
   * set, throws an exception.
   */
  std::string
  getLastAppEventRecordAsJSON(EventSeverity minimumSeverity = EventSevNone,
                              const std::string &eventRecordPath = "");

  /**
   * Return a map describing the last recorded
   * application event.  If there is no recorded event, or the
   * event has a \ref EventSeverity below \ref minimumSeverity,
   * an empty map is returned.
   *
   * Looks in the path specified by \param eventRecordPath
   * for an event record.  If this is empty, falls back to
   * the path in `IPU_APP_EVENT_RECORD_PATH`.  If neither are
   * set, throws an exception.
   */
  std::map<std::string, std::string>
  getLastAppEventRecord(EventSeverity minimumSeverity = EventSevNone,
                        const std::string &eventRecordPath = "");

  /**
   * Return the \ref EventSeverity of the last event in the event record.
   * If there is no recorded event,  \ref EventSevNone is returned.
   *
   * Looks in the path specified by \param eventRecordPath
   * for an event record.  If this is empty, falls back to
   * the path in `IPU_APP_EVENT_RECORD_PATH`.  If neither are
   * set, throws an exception.
   */
  EventSeverity
  getLastAppEventRecordSeverity(const std::string &eventRecordPath = "");
  ///@}

  ///@{
  /**
   * @name Device health check methods.
   */
  /**
   * Run basic 'health checks' on all currently configured IPUs.
   *
   * If all devices appear to be operating normally, returns an empty JSON
   * object:
   * \code
   * {}
   * \endcode
   * If any malfunctioning devices were discovered, returns
   * a JSON tree idenfifying the affected IPUs and the IPU-Machine
   * host and partition they belong to.
   * e.g.
   * \code
   * {
   *  "hosts": {
   *    "10.1.5.10": [
   *      {
   *        "error": "device",
   *        "id": "2",
   *        "partition": "p1"
   *      },
   *      {
   *        "error": "connection",
   *        "id": "3",
   *        "partition": "p1"
   *      }
   *    ]
   *  }
   * }
   * \endcode
   * There are two error types defined:
   * -  "connection" - the client was unable to communicate with the IPUoF
   * server (either because of network issues or server error) within the
   * specified timeout.
   * - "device" - the IPUoF server discovered a problem with the IPU or
   * RNIC device.
   *
   * Each IPU health check must complete within \ref timeoutMS,
   * or else a "connection" error will be recorded.
   *
   * By default, devices which are currently in use by applications
   * are not checked, unless \ref checkActiveIPUs is true.
   */
  std::string checkHealthOfDevices(unsigned timeoutMS,
                                   bool checkActiveIPUs = false);
  ///@}

  /// @private
  static int __unused;
};

#endif // _GCIPUINFO_H
