package gcipuinfo

// #cgo LDFLAGS: -ldl
// #include <dlfcn.h>
// #include <stdbool.h>
// typedef const char* (*gcipuinfo_initialise_p)(unsigned);
//
// const char* bridge_gcipuinfo_initialise(gcipuinfo_initialise_p f, unsigned discoveryMode) {
//  return f(discoveryMode);
//}
//
// typedef const char* (*gcipuinfo_get_devices_as_json_p)();
//
// const char* bridge_gcipuinfo_get_devices_as_json(gcipuinfo_get_devices_as_json_p f) {
//  return f();
//}
//
// typedef const char* (*gcipuinfo_get_last_application_event_record_as_json_p)(unsigned, const char *);
//
// const char* bridge_gcipuinfo_get_last_application_event_record_as_json(gcipuinfo_get_last_application_event_record_as_json_p f, unsigned severity, const char *eventRecordPath) {
//  return f(severity, eventRecordPath);
//}
//
// typedef unsigned (*gcipuinfo_get_last_application_event_severity_p)(const char *);
//
// unsigned bridge_gcipuinfo_get_last_application_event_severity(gcipuinfo_get_last_application_event_severity_p f, const char *eventRecordPath) {
//  return f(eventRecordPath);
//}
//
// typedef const char* (*gcipuinfo_check_health_of_devices_p)(unsigned, bool);
//
// const char* bridge_gcipuinfo_check_health_of_devices(gcipuinfo_check_health_of_devices_p f, unsigned timeoutInMS, bool checkActiveIPUs) {
//  return f(timeoutInMS, checkActiveIPUs);
//}
import "C"

import (
	"encoding/json"
	"fmt"
	"strconv"
	"strings"
	"sync"
	"unsafe"
)

// Gcipuinfo is the main inventory object which stores data
type Gcipuinfo struct {
	jsonString                    string
	data                          []interface{}
	autoUpdate                    bool
	getDevicesAsJSON              unsafe.Pointer
	getLastAppEventRecordAsJSON   unsafe.Pointer
	getLastAppEventRecordSeverity unsafe.Pointer
	checkHealthOfDevices          unsafe.Pointer
	mutex                         *sync.RWMutex
}

const (
	EventSevNone                = 0
	EventSevWarning             = 1
	EventSevApplicationError    = 2
	EventSevUndeterminedError   = 3
	EventSevRequiresUserReset   = 4
	EventSevRequiresSystemReset = 5
	EventSevNonRecoverable      = 6
)

const (
	KeyRecordPath       = "event record path"
	KeyTimestamp        = "timestamp"
	KeySeverity         = "severity"
	KeyCommandLine      = "command line"
	KeyPid              = "pid"
	KeyAttachedIPUs     = "attached ipus"
	KeySpecificIPUs     = "specific ipus"
	KeyAttachedIPUHosts = "attached ipu hosts"
	KeySpecificIPUHosts = "specific ipu hosts"
	KeyPartition        = "partition"
	KeyDescription      = "description"
)

const (
	DiscoverActivePartitionIPUs = 0
	DiscoverAllPartitionIPUs = 1
)

func NewGcipuinfo(args ...int) Gcipuinfo {
	discoveryMode := DiscoverActivePartitionIPUs
	if len(args) > 0 {
		discoveryMode = args[0]
	}
	inventory := Gcipuinfo{"[]", nil, false, nil, nil, nil, nil, &sync.RWMutex{}}
	p := &inventory
	handle := C.dlopen(C.CString("libgcipuinfo_ext.so"), C.RTLD_LAZY)
	if handle == nil {
		panic("Unable to open libgcipuinfo_ext.so")
	}
	initialise := C.dlsym(handle, C.CString("gcipuinfo_initialise"))
	if initialise == nil {
		panic("Unable to find gcipuinfo_initialise")
	}

	fptr := C.dlsym(handle, C.CString("gcipuinfo_get_devices_as_json"))
	if fptr == nil {
		panic("Unable to find gcipuinfo_get_devices_as_json")
	}
	inventory.getDevicesAsJSON = fptr

	fptr = C.dlsym(handle, C.CString("gcipuinfo_get_last_application_event_record_as_json"))
	if fptr == nil {
		panic("Unable to find gcipuinfo_get_last_application_event_record_as_json")
	}
	inventory.getLastAppEventRecordAsJSON = fptr

	fptr = C.dlsym(handle, C.CString("gcipuinfo_get_last_application_event_severity"))
	if fptr == nil {
		panic("Unable to find gcipuinfo_get_last_application_event_severity")
	}
	inventory.getLastAppEventRecordSeverity = fptr

	fptr = C.dlsym(handle, C.CString("gcipuinfo_check_health_of_devices"))
	if fptr == nil {
		panic("Unable to find gcipuinfo_check_health_of_devices")
	}
	inventory.checkHealthOfDevices = fptr

	C.bridge_gcipuinfo_initialise(C.gcipuinfo_initialise_p(initialise), C.uint(discoveryMode))

	p.UpdateData()
	return inventory
}

func (inventory *Gcipuinfo) SetUpdateMode(autoUpdate bool) {
	inventory.autoUpdate = autoUpdate
}

func (inventory *Gcipuinfo) UpdateData() {
	inventory.mutex.Lock()
	defer inventory.mutex.Unlock()

	inventory.jsonString = C.GoString(C.bridge_gcipuinfo_get_devices_as_json(C.gcipuinfo_get_devices_as_json_p(inventory.getDevicesAsJSON)))
	var temp map[string]interface{}
	json.Unmarshal([]byte(inventory.jsonString), &temp)
	inventory.data = make([]interface{}, len(temp))
	for k, v := range temp {
		var deviceId int
		deviceId, err := strconv.Atoi(k)
		if err != nil {
			panic(fmt.Sprintf("Unable to convert %v to device id", k))
		}
		inventory.data[deviceId] = v
	}
}

func (inventory *Gcipuinfo) GetAttributesForDevice(deviceId int) map[string]interface{} {
	inventory.mutex.RLock()
	defer inventory.mutex.RUnlock()

	if inventory.autoUpdate {
		inventory.UpdateData()
	}
	return inventory.data[deviceId].(map[string]interface{})
}

func (inventory *Gcipuinfo) GetNamedAttributeForDevice(deviceId int, key string) string {
	if inventory.autoUpdate {
		inventory.UpdateData()
	}
	val := inventory.GetAttributesForDevice(deviceId)[key]
	if val != nil {
		return val.(string)
	}
	return ""
}

func (inventory *Gcipuinfo) GetNamedAttributeForAll(key string) []string {
	inventory.mutex.RLock()
	defer inventory.mutex.RUnlock()

	if inventory.autoUpdate {
		inventory.UpdateData()
	}
	var a []string
	for _, deviceAttrs := range inventory.data {
		val := deviceAttrs.(map[string]interface{})[key]
		if val != nil {
			a = append(a, val.(string))
		}
	}
	return a
}

func (inventory *Gcipuinfo) GetDevices() []interface{} {
	inventory.mutex.RLock()
	defer inventory.mutex.RUnlock()

	if inventory.autoUpdate {
		inventory.UpdateData()
	}
	return inventory.data
}

func (inventory *Gcipuinfo) GetDevicesAsJSON() string {
	inventory.mutex.RLock()
	defer inventory.mutex.RUnlock()

	if inventory.autoUpdate {
		inventory.UpdateData()
	}
	return inventory.jsonString
}

func (inventory *Gcipuinfo) GetLastAppEventRecordAsJSON(severity uint, eventRecordPath string) string {
	return C.GoString(C.bridge_gcipuinfo_get_last_application_event_record_as_json(C.gcipuinfo_get_last_application_event_record_as_json_p(inventory.getLastAppEventRecordAsJSON), C.uint(severity), C.CString(eventRecordPath)))
}

func (inventory *Gcipuinfo) GetLastAppEventRecord(severity uint, eventRecordPath string) map[string]interface{} {
	jsonEventRecord := inventory.GetLastAppEventRecordAsJSON(severity, eventRecordPath)
	var temp map[string]interface{}

	dec := json.NewDecoder(strings.NewReader(jsonEventRecord))
	dec.UseNumber()
	err := dec.Decode(&temp)
	if err != nil {
		panic(fmt.Sprintf("Unable to decode json event record"))
	}

	return temp
}

func (inventory *Gcipuinfo) GetLastAppEventRecordSeverity(eventRecordPath string) uint {
	return uint(C.bridge_gcipuinfo_get_last_application_event_severity(C.gcipuinfo_get_last_application_event_severity_p(inventory.getLastAppEventRecordSeverity), C.CString(eventRecordPath)))
}

func (inventory *Gcipuinfo) CheckHealthOfDevices(timeoutMS uint, checkActiveIPUs bool) string {
	return C.GoString(C.bridge_gcipuinfo_check_health_of_devices(C.gcipuinfo_check_health_of_devices_p(inventory.checkHealthOfDevices), C.uint(timeoutMS), C.bool(checkActiveIPUs)))
}
