/*
 * Copyright (c) 2021 Graphcore Ltd, All rights reserved.
 */

// gcipuinfo examples:
// Check each of the application event record paths specified
// as command line arguments.  If an entry is found, display it.
// Also collect a list of all IPU-Machine hosts named in
// at least one event.

#include <iostream>
#include <set>

#include "graphcore_target_access/gcipuinfo/gcipuinfo.h"
#include <nlohmann/json.hpp>
using json = nlohmann::json;

int main(int argc, char *argv[]) {

  if (argc < 2) {
    std::cerr << __FILE__ << ": error requires at least one argument\n";
    return 1;
  }

  gcipuinfo inventory;

  std::set<std::string> hostsNamedInErrors;

  for (int i = 1; i < argc; i++) {
    std::cout << "== Checking application event record: " << argv[i] << "\n";

    std::string eventRecordJSON =
        inventory.getLastAppEventRecordAsJSON(gcipuinfo::EventSevNone, argv[i]);

    json j = json::parse(eventRecordJSON);
    if (j.empty()) {
      std::cout << "No event record entry found.\n";
    } else {
      std::cout << j.dump(2) << "\n";

      // If the event has named specific IPUs
      // record their associated machine names
      bool foundSpecificHosts = false;
      if (j.contains(gcipuinfo::keySpecificIPUHosts)) {
        auto hosts = j[gcipuinfo::keySpecificIPUHosts];
        foundSpecificHosts = hosts.size() > 0;
        for (const auto &element : hosts.items()) {
          hostsNamedInErrors.insert(element.value().dump());
        }
      }
      if (!foundSpecificHosts && j.contains(gcipuinfo::keyAttachedIPUHosts)) {
        // if no specific IPUs have been mentioned in this error,
        // fall back to recording hosts of all attached IPUs
        auto hosts = j[gcipuinfo::keyAttachedIPUHosts];
        for (const auto &element : hosts.items()) {
          hostsNamedInErrors.insert(element.value().dump());
        }
      }
      auto severity = inventory.getLastAppEventRecordSeverity(argv[i]);
      if (severity > gcipuinfo::EventSevWarning) {
        std::cout << "Last event was an error.\n";
      }
    }
    std::cout << "\n";
  }

  if (hostsNamedInErrors.size() > 0) {
    std::cout << "== IPU-Machine hosts involved in errors, across all "
                 "application event records:\n";
    for (auto host : hostsNamedInErrors) {
      std::cout << "  " << host << "\n";
    }
  }

  return 0;
}
