/*
 * Copyright (c) 2021 Graphcore Ltd, All rights reserved.
 *
 * This file implements an alternative way of accessing the gcipuinfo API
 * for clients that cannot link against the GCDA or Poplar
 * C++ libraries.
 * It uses the gcipuinfo_get_devices_as_json() C function
 * exported by gcipuinfo_ext.so
 *
 */

#include <dlfcn.h>
#include <iostream>
#include <nlohmann/json.hpp>
#include <sstream>

#include "graphcore_target_access/gcipuinfo/gcipuinfo.h"

using json = nlohmann::json;

typedef void (*gcipuinfo_initialise_p)(unsigned);
static gcipuinfo_initialise_p gcipuinfo_initialise = nullptr;

typedef const char *(*gcipuinfo_get_devices_as_json_p)(void);
static gcipuinfo_get_devices_as_json_p gcipuinfo_get_devices_as_json = nullptr;

typedef const char *(*gcipuinfo_get_last_application_event_record_as_json_p)(
    unsigned, const char *);
static gcipuinfo_get_last_application_event_record_as_json_p
    gcipuinfo_get_last_application_event_record_as_json = nullptr;

typedef unsigned (*gcipuinfo_get_last_application_event_severity_p)(
    const char *);
static gcipuinfo_get_last_application_event_severity_p
    gcipuinfo_get_last_application_event_severity = nullptr;

typedef const char *(*gcipuinfo_check_health_of_devices_p)(unsigned, bool);
static gcipuinfo_check_health_of_devices_p gcipuinfo_check_health_of_devices =
    nullptr;

static bool s_initialised = false;

gcipuinfo::gcipuinfo(DeviceDiscoveryMode discoveryMode) {

  if (!s_initialised) {
    void *handle = dlopen("libgcipuinfo_ext.so", RTLD_LOCAL | RTLD_LAZY);
    if (!handle) {
      std::cerr << dlerror() << "\n";
      throw std::runtime_error("Unable to open libgcipuinfo_ext.so");
    }

    void *fptr = dlsym(handle, "gcipuinfo_initialise");
    if (fptr == nullptr) {
      throw std::runtime_error("Unable to find gcipuinfo_initialise");
    }
    gcipuinfo_initialise = reinterpret_cast<gcipuinfo_initialise_p>(fptr);

    fptr = dlsym(handle, "gcipuinfo_get_devices_as_json");
    if (fptr == nullptr) {
      throw std::runtime_error("Unable to find gcipuinfo_get_devices_as_json");
    }
    gcipuinfo_get_devices_as_json =
        reinterpret_cast<gcipuinfo_get_devices_as_json_p>(fptr);

    fptr = dlsym(handle, "gcipuinfo_get_last_application_event_record_as_json");
    if (fptr == nullptr) {
      throw std::runtime_error(
          "Unable to find gcipuinfo_get_last_application_event_record_as_json");
    }
    gcipuinfo_get_last_application_event_record_as_json =
        reinterpret_cast<gcipuinfo_get_last_application_event_record_as_json_p>(
            fptr);

    fptr = dlsym(handle, "gcipuinfo_get_last_application_event_severity");
    if (fptr == nullptr) {
      throw std::runtime_error(
          "Unable to find gcipuinfo_get_last_application_event_severity");
    }
    gcipuinfo_get_last_application_event_severity =
        reinterpret_cast<gcipuinfo_get_last_application_event_severity_p>(fptr);

    fptr = dlsym(handle, "gcipuinfo_check_health_of_devices");
    if (fptr == nullptr) {
      throw std::runtime_error(
          "Unable to find gcipuinfo_check_health_of_devices");
    }
    gcipuinfo_check_health_of_devices =
        reinterpret_cast<gcipuinfo_check_health_of_devices_p>(fptr);

    s_initialised = true;
  }
  gcipuinfo_initialise(discoveryMode);

  if (!updateData()) {
    throw std::runtime_error("Failed to update data");
  }
}

bool gcipuinfo::updateData() {
  data.clear();

  const char *output = gcipuinfo_get_devices_as_json();

  json j;

  if (!json::accept(output)) {
    throw std::runtime_error("Unable to parse json");
  }

  j = json::parse(output);

  data.clear();
  data.resize(j.size());

  unsigned index = 0;
  for (auto &device : j) {
    for (auto &item : device.items()) {
      data[index][item.key()] = item.value();
    }
    index++;
  }

  return true;
}

std::string gcipuinfo::getDevicesAsJSON() {
  return gcipuinfo_get_devices_as_json();
}

void gcipuinfo::setUpdateMode(bool autoUpdate) { autoUpdateMode = autoUpdate; }

std::vector<std::map<std::string, std::string>> gcipuinfo::getDevices() {
  if (autoUpdateMode) {
    updateData();
  }
  return data;
}

std::map<std::string, std::string>
gcipuinfo::getAttributesForDevice(unsigned deviceId) {
  if (autoUpdateMode) {
    updateData();
  }
  if (deviceId < data.size()) {
    return data[deviceId];
  }
  return {};
}

std::string gcipuinfo::getNamedAttributeForDevice(unsigned deviceId,
                                                  const std::string key) {
  if (autoUpdateMode) {
    updateData();
  }

  if (deviceId < data.size()) {
    auto &dict = data[deviceId];
    auto search = dict.find(key);
    if (search != dict.end()) {
      return search->second;
    }
  }
  return "";
}

std::vector<std::string>
gcipuinfo::getNamedAttributeForAll(const std::string key) {
  if (autoUpdateMode) {
    updateData();
  }

  std::vector<std::string> result;
  for (auto &device : data) {
    if (device.count(key) > 0) {
      result.push_back(device[key]);
    }
  }
  return result;
}

std::string
gcipuinfo::getLastAppEventRecordAsJSON(gcipuinfo::EventSeverity minimumSeverity,
                                       const std::string &eventRecordPath) {
  return gcipuinfo_get_last_application_event_record_as_json(
      static_cast<unsigned>(minimumSeverity), eventRecordPath.c_str());
}

std::map<std::string, std::string>
gcipuinfo::getLastAppEventRecord(gcipuinfo::EventSeverity minimumSeverity,
                                 const std::string &eventRecordPath) {

  std::map<std::string, std::string> map;
  nlohmann::ordered_json eventRecord = nlohmann::ordered_json::parse(
      getLastAppEventRecordAsJSON(minimumSeverity, eventRecordPath));

  for (const auto &element : eventRecord.items()) {
    map[element.key()] = element.value().dump();
  }

  return map;
}

gcipuinfo::EventSeverity
gcipuinfo::getLastAppEventRecordSeverity(const std::string &eventRecordPath) {
  return static_cast<gcipuinfo::EventSeverity>(
      gcipuinfo_get_last_application_event_severity(eventRecordPath.c_str()));
}

std::string gcipuinfo::checkHealthOfDevices(unsigned timeoutMS,
                                            bool checkActiveIPUs) {
  return gcipuinfo_check_health_of_devices(timeoutMS, checkActiveIPUs);
}
